#include "wresample.hh"
#include "matlab_fcs.hh"
#include "lcl_array.hh"
#include "Bits.hh"
//#include "DecimateBy2.hh"
#include "Hanning.hh"
#include "DVecType.hh"
#include "FIRdft.hh"
#include "FIRdesign.hh"
#include <iostream>

//--------------------------------------------------------------------------
//
// The FIRDFT_ZERO_PHASE macro causes the FIRdft filter to run in zero 
// phase shift mode. This means that the output time series start time 
// is adjusted to cancel the filter phase delay. Using this flag should
// have no effect on the results and minimal effect on the speed and
// complexity of the current code, but it should help simplify any future
// modification to implement data streaming.
//
#define FIRDFT_ZERO_PHASE 1
//--------------------------------------------------------------------------

using namespace std;

namespace wpipe {


  //======================================  Resampler contructor
  resampler::resampler(void) {}

  //======================================  Resampler destructor
  resampler::~resampler(void) {
    for (filt_iter i=rate_map.begin(); i != rate_map.end(); ++i) {
      delete i->second;
      i->second = 0;
    }
    rate_map.clear();
  }

  //======================================  Resample function
  tser_vect
  resampler::wresample(const tser_vect& data, double sampleFrequency) {
    int numberOfChannels = data.size();

    ///////////////////////////////////////////////////////////////////////////
    //                   validate command line arguments                     //
    ///////////////////////////////////////////////////////////////////////////
 
    //==================================   validate equal data durations
    Interval dataDuration = data[0].getInterval();
    for (int chanNumber=1; chanNumber < numberOfChannels; chanNumber++) {
      if (dataDuration != data[chanNumber].getInterval()) {
	error("data durations are not equal");
      }
    }

    //////////////////////////////////////////////////////////////////////////
    //                            resample data                             //
    //////////////////////////////////////////////////////////////////////////

    // begin loop over channels
    tser_vect ts(numberOfChannels);
    long fSample = long(sampleFrequency);
    for (int chanNumber=0; chanNumber<numberOfChannels; chanNumber++){
      TSeries tsin(data[chanNumber]);
      tsin.Convert(DVector::t_double);

      // for now demand reduction by a power of 2
      double decimFact = sampleFrequency*double(tsin.getTStep());
      if (decimFact == 0) {
	error("Decimation factor is zero");
      }
      decimFact = 1.0/decimFact;
      if (!is_power_of_2(int(decimFact))) {
	cerr << "Target rate: " << sampleFrequency << " input t-step: " 
	     << tsin.getTStep() << endl;
	error("Sample rate reduction not a power of 2");
      }

      long fRawSample = long(1./double(tsin.getTStep()));
      if (fSample != fRawSample) {
	// resample data
	//ts = resample(data, upSampleFactor, downSampleFactor, filter)

	//------------------------------  Get an anti-aliasing filter 
	//                                Note: assume integer sampling rates
	FIRdft* filt = dynamic_cast<FIRdft*>(filter(fRawSample, fSample));
	Interval tShift = filt->getTimeDelay();

	//------------------------------  Filter and decimate
	long downSampleFactor = fRawSample / fSample;
	tsin=(*filt)(tsin).decimate(downSampleFactor);

	//  remove data from before nominal start time.
#ifndef FIRDFT_ZERO_PHASE
	long nShift = long(tShift / tsin.getTStep());
	tsin.refDVect()->replace_with_zeros(0, nShift, 0);
#else
	tsin.eraseStart(tShift);
#endif
	tsin.extend(tsin.getEndTime() + tShift);
	delete filt;
      }
      ts[chanNumber] = tsin;
    }      // end loop over channels

    ///////////////////////////////////////////////////////////////////////////
    //                                  return                               //
    ///////////////////////////////////////////////////////////////////////////

    // return to calling function
    return ts;
  }

  Pipe*
  resampler::filter(long fRawSample, long fSample) {
    rate_pair rp(fRawSample, fSample);
    if (rate_map.find(rp) == rate_map.end()) {
      long denom      = gcd(fSample, fRawSample);
      long upSampleFactor   = fSample /denom;
      long downSampleFactor = fRawSample / denom; 

      // design anti-alias filter
      const int nBand = 2;
      int filterOrder = 2 * 256 * max(upSampleFactor, downSampleFactor);
      double filterCutoff = 0.99 / max(upSampleFactor, downSampleFactor);
      //int filterOrder = 2 * 16 * max(upSampleFactor, downSampleFactor);
      //double filterCutoff = 0.90 / max(upSampleFactor, downSampleFactor);
      double filterFrequencies[2*nBand] = {0.0, filterCutoff, filterCutoff, 1.0};
      double filterMagnitudes[2*nBand]  = {1.0, 1.0, 0.0, 0.0};
      double wband[nBand] = {1.0, 1.0};
      DVectD filter_coefs(filterOrder+1);
      firls(filterOrder, nBand, filterFrequencies, filterMagnitudes, wband, 
	    filter_coefs.refTData());
      Hanning filter1(filterOrder + 1);
      filter_coefs *= filter1.refDVect();
      filter_coefs *= double(upSampleFactor)
	             /filter1.refDVect().getDouble(filterOrder/2);
      FIRdft* filt = new FIRdft(filterOrder, double(fRawSample));
      filt->setCoefs(filterOrder+1, filter_coefs.refTData());
#ifdef FIRDFT_ZERO_PHASE
      filt->setMode(FIRdft::fm_zero_phase);
#endif
      rate_map.insert(filt_node(rp, filt));
    }
    return rate_map.find(rp)->second->clone();
  }

}  // namespace wpipe
